#!/usr/bin/python
'''
    Title: img2bbc
    Author: TheOuterLinux
    Purpose: To make it easier to convert images to DATA array for
             BBC Micro and Matrix Brandy BASIC V projects. 
             Exports as FILENAME.DAT.
             
    Last updated: 2020/07/01
             
    Copyright (C) 2020  TheOuterLinux

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.

    img2bbc.py requirements:
    ----------------------
        Python2.7 or Python3
        ImageMagick
        PIL (Python 2.7 - 3.5 via pip) or python3-pillowfight (Python 3.6 +)
        numpy
        ptyprocess
        six
        colorama
        
'''
import os
from os import system
import sys
import signal
from PIL import Image
import subprocess
import numpy as np
from subprocess import call
import tempfile
from tempfile import gettempdir
from shutil import rmtree
from six.moves import input as raw_input
from colorama import init, Fore, Back, Style

import signal
import time
import sys

def clear(): 
    # check and make call for specific operating system 
    _ = call('clear' if os.name =='posix' else 'cls') 

#Filter ANSI escape sequences out of any text sent to stdout or stderr, 
#and replace them with equivalent Win32 calls if using Windows.
if os.name == 'Windows': init()

#Create temporary directory for palettes.
#This method makes things much more cross-platform.
tmp = os.path.join(gettempdir(), '.{}'.format(hash(os.times())))
os.makedirs(tmp)
DIR = os.path.dirname(os.path.realpath(__file__))

#2 Colors (Monochrome)
palette2 = DIR+'/Palettes/img2bbc/palette2.gif'

#4 Colors
palette4 = DIR+'/Palettes/img2bbc/palette4.gif'

#8 Colors
palette8 = DIR+'/Palettes/img2bbc/palette8.gif'

#256 Colors
palette32 = DIR+'/Palettes/img2bbc/palette32.gif'

#Clear the terminal screen
clear()

#Title
if os.name == 'Windows':
	myCoolTitle='img2bbc.py by TheOuterLinux'
	system('title '+myCoolTitle)
else:
	 sys.stdout.write('\x1b]2;img2bbc.py by TheOuterLinux\x07')
print (' ' + Fore.CYAN + Back.CYAN + '+-------------------------------------------------------------------++' + Style.RESET_ALL)
print (' ' + Fore.CYAN + Back.CYAN + '||'+ Fore.MAGENTA + Back.WHITE + ' img2bbc.py - Convert a modern image to BBC BASIC DATA array      ' + Fore.CYAN + Back.CYAN + '||' + Style.RESET_ALL)
print (' ' + Fore.CYAN + Back.CYAN + '||'+ Fore.MAGENTA + Back.WHITE + ' Author - TheOuterLinux (https://theouterlinux.gitlab.io)         ' + Fore.CYAN + Back.CYAN + '||' + Style.RESET_ALL)
print (' ' + Fore.CYAN + Back.CYAN + '||'+ Fore.MAGENTA + Back.WHITE + ' License - GPL v3 or newer (https://www.gnu.org/licenses/)        ' + Fore.CYAN + Back.CYAN + '||' + Style.RESET_ALL)
#print (' ' + Fore.CYAN + Back.CYAN + '+-------------------------------------------------------------------++' + Style.RESET_ALL)
print (' ' + Fore.CYAN + Back.CYAN + "||"+ Fore.BLACK + Back.CYAN + "           Use 'q' or 'quit' any time to end program.             " + Fore.CYAN + Back.CYAN + "||" + Style.RESET_ALL)
print (' ' + Fore.CYAN + Back.CYAN + ' ' + Back.BLACK)

#Print temporarily created palette locations
print(' ' + Fore.CYAN + Back.CYAN + ' ' + Back.BLACK + ' Temporary folder created in ' + Fore.WHITE + tmp + Fore.CYAN + '.')
print(' ' + Fore.CYAN + Back.CYAN + ' ' + Back.BLACK + '')

# Ask for the full path of the file to convert. The xfce4-terminal
# browser supports drag-and-drop.
filepath = raw_input(' ' + Fore.CYAN + Back.CYAN + ' ' + Back.BLACK + ' File to convert (full path): ' + Style.RESET_ALL)
filepath = filepath.replace("'", '')
filepath = filepath.strip()
    
if filepath in ['q','quit','Q','Quit','QUIT','Exit','EXIT','exit']:
	rmtree(tmp, ignore_errors=True)
	print('')
	exit()
if os.path.exists(filepath):
	print('')
else:
	clear()
	print (' ' + Fore.CYAN + Back.CYAN + '+-------------------------------------------------------------------++' + Style.RESET_ALL)
	print (' ' + Fore.CYAN + Back.CYAN + '||'+ Fore.MAGENTA + Back.WHITE + ' img2bbc.py - Convert a modern image to BBC BASIC DATA array      ' + Fore.CYAN + Back.CYAN + '||' + Style.RESET_ALL)
	print (' ' + Fore.CYAN + Back.CYAN + '||'+ Fore.MAGENTA + Back.WHITE + ' Author - TheOuterLinux (https://theouterlinux.gitlab.io)         ' + Fore.CYAN + Back.CYAN + '||' + Style.RESET_ALL)
	print (' ' + Fore.CYAN + Back.CYAN + '||'+ Fore.MAGENTA + Back.WHITE + ' License - GPL v3 or newer (https://www.gnu.org/licenses/)        ' + Fore.CYAN + Back.CYAN + '||' + Style.RESET_ALL)
	#print (' ' + Fore.CYAN + Back.CYAN + '+-------------------------------------------------------------------++' + Style.RESET_ALL)
	print (' ' + Fore.CYAN + Back.CYAN + "||"+ Fore.MAGENTA + Back.CYAN + "           Use 'q' or 'quit' any time to end program.             " + Fore.CYAN + Back.CYAN + "||" + Style.RESET_ALL)
	print (' ' + Fore.CYAN + Back.CYAN + ' ' + Back.BLACK + '                                                                    ' + Back.CYAN + ' ' + Style.RESET_ALL)
	print (' ' + Fore.CYAN + Back.CYAN + ' ' + Back.BLACK + ' You must input an existing full file path to use this program.     ' + Back.CYAN + ' ' + Style.RESET_ALL)
	print (' ' + Fore.CYAN + Back.CYAN + ' ' + Back.BLACK + ' Please use one of the following formats:                           ' + Back.CYAN + ' ' + Style.RESET_ALL)
	print (' ' + Fore.CYAN + Back.CYAN + ' ' + Back.BLACK + ' ----------------------------------------                           ' + Back.CYAN + ' ' + Style.RESET_ALL)
	print (' ' + Fore.CYAN + Back.CYAN + ' ' + Back.BLACK + '     *.blp                 *.pcx                                    ' + Back.CYAN + ' ' + Style.RESET_ALL)
	print (' ' + Fore.CYAN + Back.CYAN + ' ' + Back.BLACK + '     *.bmp                 *.pdf                                    ' + Back.CYAN + ' ' + Style.RESET_ALL)
	print (' ' + Fore.CYAN + Back.CYAN + ' ' + Back.BLACK + '     *.cur                 *.png                                    ' + Back.CYAN + ' ' + Style.RESET_ALL)
	print (' ' + Fore.CYAN + Back.CYAN + ' ' + Back.BLACK + '     *.gbr                 *.ppm                                    ' + Back.CYAN + ' ' + Style.RESET_ALL)
	print (' ' + Fore.CYAN + Back.CYAN + ' ' + Back.BLACK + '     *.gif                 *.psd                                    ' + Back.CYAN + ' ' + Style.RESET_ALL)
	print (' ' + Fore.CYAN + Back.CYAN + ' ' + Back.BLACK + '     *.icns                *.tga                                    ' + Back.CYAN + ' ' + Style.RESET_ALL)
	print (' ' + Fore.CYAN + Back.CYAN + ' ' + Back.BLACK + '     *.ico                 *.tif, tiff                              ' + Back.CYAN + ' ' + Style.RESET_ALL)
	print (' ' + Fore.CYAN + Back.CYAN + ' ' + Back.BLACK + '     *.jp2 (JPEG 2000)     *.wal                                    ' + Back.CYAN + ' ' + Style.RESET_ALL)
	print (' ' + Fore.CYAN + Back.CYAN + ' ' + Back.BLACK + '     *.jpg, jpeg           *.webp                                   ' + Back.CYAN + ' ' + Style.RESET_ALL)
	print (' ' + Fore.CYAN + Back.CYAN + ' ' + Back.BLACK + '     *.mic                 *.xbm                                    ' + Back.CYAN + ' ' + Style.RESET_ALL)
	print (' ' + Fore.CYAN + Back.CYAN + ' ' + Back.BLACK + '     *.mps                 *.xpm                                    ' + Back.CYAN + ' ' + Style.RESET_ALL)
	print (' ' + Fore.CYAN + Back.CYAN + ' ' + Back.BLACK + '     *.palm                *.xv                                     ' + Back.CYAN + ' ' + Style.RESET_ALL)
	print (' ' + Fore.CYAN + Back.CYAN + ' ' + Back.BLACK + '                                                                    ' + Back.CYAN + ' ' + Style.RESET_ALL)
	print (' ' + Fore.CYAN + Back.CYAN + '                                                                      ' + Back.BLACK + ' ')
	print ('' + Style.RESET_ALL)
	rmtree(tmp, ignore_errors=True)
	exit()

# Convert file extensions to all lower-case since older file systems
# tend to use all upper-case while newer file systems can use upper and
# lower-case. Avoids the hassle of writing more if-than statements than
# necessary.
fileext = os.path.splitext(filepath.lower())[1]
extcount = len(fileext)

# Get the file name without the full path or extension (in most cases)
base = os.path.basename(filepath)
base = os.path.splitext(base)[0]

# Get the path of the directory the original file is in so we know where 
# to put the converted result.
dirpath = os.path.dirname(filepath)

# The file name and path to use for a preview file
previewfile = dirpath + '/' + base + '_img2bbc_preview.gif'

# Prepare the BBC BASIC file name to make it DOS-friendly by removing
# spaces and shortening the file name to 8 characters. Using 'DAT' as
# the file extension to stand for 'BBC BASIC DATA'. And because it is
# still plain text, it shouldn't trick the computer into thinking it is
# a any sort of special file-type.
dosbase = base.replace(' ', '')
dosbase = dosbase[:8]
dosbase = dosbase.upper()
bbcfilepath = tmp + '/' + dosbase + '.DAT'
bbcfilepathfinal = dirpath + '/' + dosbase + '.DAT'

# The PIL library supports more formats than this, but only supporting
# the following makes the most sense.
if  fileext in ['.blp', '.bmp', '.cur', '.gbr', '.gif', '.icns', '.ico', \
                '.jp2', '.jpeg', '.jpg', '.mic', '.msp', '.palm', '.pcx', \
                '.pdf', '.png', '.ppm', '.psd', '.tga', '.tif', '.tiff', \
                '.wal', '.webp', '.xbm', '.xpm', '.xv']:
	print (' ' + Fore.CYAN + Back.CYAN + ' ' + Back.BLACK + ' File format ' + Fore.WHITE + fileext + Fore.CYAN + ' supported.')
	print (' ' + Fore.CYAN + Back.CYAN + ' ' + Back.BLACK + '')
else:
	clear()
	print (' ' + Fore.CYAN + Back.CYAN + '+-------------------------------------------------------------------++' + Style.RESET_ALL)
	print (' ' + Fore.CYAN + Back.CYAN + '||'+ Fore.MAGENTA + Back.WHITE + ' img2bbc.py - Convert a modern image to BBC BASIC DATA array      ' + Fore.CYAN + Back.CYAN + '||' + Style.RESET_ALL)
	print (' ' + Fore.CYAN + Back.CYAN + '||'+ Fore.MAGENTA + Back.WHITE + ' Author - TheOuterLinux (https://theouterlinux.gitlab.io)         ' + Fore.CYAN + Back.CYAN + '||' + Style.RESET_ALL)
	print (' ' + Fore.CYAN + Back.CYAN + '||'+ Fore.MAGENTA + Back.WHITE + ' License - GPL v3 or newer (https://www.gnu.org/licenses/)        ' + Fore.CYAN + Back.CYAN + '||' + Style.RESET_ALL)
	#print (' ' + Fore.CYAN + Back.CYAN + '+-------------------------------------------------------------------++' + Style.RESET_ALL)
	print (' ' + Fore.CYAN + Back.CYAN + "||"+ Fore.MAGENTA + Back.CYAN + "           Use 'q' or 'quit' any time to end program.             " + Fore.CYAN + Back.CYAN + "||" + Style.RESET_ALL)
	print (' ' + Fore.CYAN + Back.CYAN + ' ' + Back.BLACK + '                                                                    ' + Back.CYAN + ' ' + Style.RESET_ALL)
	if extcount == 0: print (' ' + Fore.CYAN + Back.CYAN + ' ' + Back.BLACK + ' You must input an existing full file path to use this program.     ' + Back.CYAN + ' ' + Style.RESET_ALL)
	if extcount == 1: print (' ' + Fore.CYAN + Back.CYAN + ' ' + Back.BLACK + ' File format *' + fileext + ' not supported.                                      ' + Back.CYAN + ' ' + Style.RESET_ALL)
	if extcount == 2: print (' ' + Fore.CYAN + Back.CYAN + ' ' + Back.BLACK + ' File format *' + fileext + ' not supported.                                     ' + Back.CYAN + ' ' + Style.RESET_ALL)
	if extcount == 3: print (' ' + Fore.CYAN + Back.CYAN + ' ' + Back.BLACK + ' File format *' + fileext + ' not supported.                                    ' + Back.CYAN + ' ' + Style.RESET_ALL)
	if extcount == 4: print (' ' + Fore.CYAN + Back.CYAN + ' ' + Back.BLACK + ' File format *' + fileext + ' not supported.                                   ' + Back.CYAN + ' ' + Style.RESET_ALL)
	if extcount == 5: print (' ' + Fore.CYAN + Back.CYAN + ' ' + Back.BLACK + ' File format *' + fileext + ' not supported.                                  ' + Back.CYAN + ' ' + Style.RESET_ALL)
	print (' ' + Fore.CYAN + Back.CYAN + ' ' + Back.BLACK + ' Please use one of the following formats:                           ' + Back.CYAN + ' ' + Style.RESET_ALL)
	print (' ' + Fore.CYAN + Back.CYAN + ' ' + Back.BLACK + ' ----------------------------------------                           ' + Back.CYAN + ' ' + Style.RESET_ALL)
	print (' ' + Fore.CYAN + Back.CYAN + ' ' + Back.BLACK + '     *.blp                 *.pcx                                    ' + Back.CYAN + ' ' + Style.RESET_ALL)
	print (' ' + Fore.CYAN + Back.CYAN + ' ' + Back.BLACK + '     *.bmp                 *.pdf                                    ' + Back.CYAN + ' ' + Style.RESET_ALL)
	print (' ' + Fore.CYAN + Back.CYAN + ' ' + Back.BLACK + '     *.cur                 *.png                                    ' + Back.CYAN + ' ' + Style.RESET_ALL)
	print (' ' + Fore.CYAN + Back.CYAN + ' ' + Back.BLACK + '     *.gbr                 *.ppm                                    ' + Back.CYAN + ' ' + Style.RESET_ALL)
	print (' ' + Fore.CYAN + Back.CYAN + ' ' + Back.BLACK + '     *.gif                 *.psd                                    ' + Back.CYAN + ' ' + Style.RESET_ALL)
	print (' ' + Fore.CYAN + Back.CYAN + ' ' + Back.BLACK + '     *.icns                *.tga                                    ' + Back.CYAN + ' ' + Style.RESET_ALL)
	print (' ' + Fore.CYAN + Back.CYAN + ' ' + Back.BLACK + '     *.ico                 *.tif, tiff                              ' + Back.CYAN + ' ' + Style.RESET_ALL)
	print (' ' + Fore.CYAN + Back.CYAN + ' ' + Back.BLACK + '     *.jp2 (JPEG 2000)     *.wal                                    ' + Back.CYAN + ' ' + Style.RESET_ALL)
	print (' ' + Fore.CYAN + Back.CYAN + ' ' + Back.BLACK + '     *.jpg, jpeg           *.webp                                   ' + Back.CYAN + ' ' + Style.RESET_ALL)
	print (' ' + Fore.CYAN + Back.CYAN + ' ' + Back.BLACK + '     *.mic                 *.xbm                                    ' + Back.CYAN + ' ' + Style.RESET_ALL)
	print (' ' + Fore.CYAN + Back.CYAN + ' ' + Back.BLACK + '     *.mps                 *.xpm                                    ' + Back.CYAN + ' ' + Style.RESET_ALL)
	print (' ' + Fore.CYAN + Back.CYAN + ' ' + Back.BLACK + '     *.palm                *.xv                                     ' + Back.CYAN + ' ' + Style.RESET_ALL)
	print (' ' + Fore.CYAN + Back.CYAN + ' ' + Back.BLACK + '                                                                    ' + Back.CYAN + ' ' + Style.RESET_ALL)
	print (' ' + Fore.CYAN + Back.CYAN + '                                                                      ' + Back.BLACK + ' ')
	print ('' + Style.RESET_ALL)
	rmtree(tmp, ignore_errors=True)
	exit()

# Get the selected image file's width and height
im = Image.open(filepath)
width, height = im.size
colorcount = len(list(filter(None, im.histogram())))
if sys.version_info[0] == 2:
	print (' ' + Fore.CYAN + Back.CYAN + ' ' + Back.BLACK + ' Image width:' + Fore.WHITE), width
	print (' ' + Fore.CYAN + Back.CYAN + ' ' + Back.BLACK + ' Image height:' + Fore.WHITE), height
	print (' ' + Fore.CYAN + Back.CYAN + ' ' + Back.BLACK + ' Image color count:' + Fore.WHITE), colorcount

# Ask what SCREEN mode will be used
print (' ' + Fore.CYAN + Back.CYAN + ' ' + Back.BLACK + '')
print (' ' + Fore.WHITE + Back.CYAN + ' ' + Back.BLACK + ' BBC Micro Modes')
print (' ' + Fore.CYAN + Back.CYAN + ' ' + Back.BLACK + ' ---------------')
print (' ' + Fore.CYAN + Back.CYAN + ' ' + Back.BLACK + ' MODE ' + Fore.WHITE + '  0' + Fore.CYAN + ' - 80x32 Text Chars 640x256 Graphics Pixels 2 Colors | Max chars per line 160')
print (' ' + Fore.CYAN + Back.CYAN + ' ' + Back.BLACK + ' MODE ' + Fore.WHITE + '  1' + Fore.CYAN + ' - 40x32 Text Chars 320x256 Graphics Pixels 4 Colors | Max chars per line 160')
print (' ' + Fore.CYAN + Back.CYAN + ' ' + Back.BLACK + ' MODE ' + Fore.WHITE + '  2' + Fore.CYAN + ' - 20x32 Text Chars 160x256 Graphics Pixels 8 Colors | Max chars per line 160')
print (' ' + Fore.CYAN + Back.CYAN + ' ' + Back.BLACK + ' MODE ' + Fore.WHITE + '  4' + Fore.CYAN + ' - 40x32 Text Chars 320x256 Graphics Pixels 2 Colors | Max chars per line 160')
print (' ' + Fore.CYAN + Back.CYAN + ' ' + Back.BLACK + ' MODE ' + Fore.WHITE + '  5' + Fore.CYAN + ' - 20x32 Text Chars 160x256 Graphics Pixels 4 Colors | Max chars per line 160')
print (' ' + Fore.CYAN + Back.CYAN + ' ' + Back.BLACK + '' + Fore.CYAN + '')
print (' ' + Fore.CYAN + Back.CYAN + ' ' + Back.BLACK + Fore.MAGENTA + '**Do not use graphics with a width higher than 75 pixels if using with an actual BBC Micro.' + Fore.CYAN + '')
print (' ' + Fore.CYAN + Back.CYAN + ' ' + Back.BLACK + '' + Fore.CYAN + '')
print (' ' + Fore.CYAN + Back.CYAN + ' ' + Fore.WHITE + Back.BLACK + ' The ' + Fore.CYAN + 'MODES' + Fore.WHITE + ' below only apply to ' + Fore.MAGENTA + 'Matrix Brandy BASIC V' + Fore.WHITE + ' (http://brandy.matrixnetwork.co.uk/) ' + Fore.CYAN + ' ')
print (' ' + Fore.CYAN + Back.CYAN + ' ' + Back.BLACK + '------------------------------------------------------------------------------------------')
print (' ' + Fore.CYAN + Back.CYAN + ' ' + Back.BLACK + Fore.CYAN + ' MODE'+ Fore.WHITE +'  15' + Fore.CYAN + ' - 80x32 Text Chars | Max chars per line 1024 | 32 Colors')
print (' ' + Fore.CYAN + Back.CYAN + ' ' + Back.BLACK + '')
modenumber = raw_input(' ' + Fore.CYAN + Back.CYAN + ' ' + Back.BLACK + ' What MODE do you plan on using? (' + Fore.WHITE + '0,1,2,..,q' + Fore.CYAN + '): ' + Style.RESET_ALL)
modenumber = modenumber.lower()
clear()
print(' ' + Fore.CYAN + Back.CYAN + '+-------------------------------------------------------------------++' + Style.RESET_ALL)
print(' ' + Fore.CYAN + Back.CYAN + '||'+ Fore.MAGENTA + Back.WHITE + ' img2bbc.py - Convert a modern image to BBC BASIC DATA array      ' + Fore.CYAN + Back.CYAN + '||' + Style.RESET_ALL)
print(' ' + Fore.CYAN + Back.CYAN + '||'+ Fore.MAGENTA + Back.WHITE + ' Author - TheOuterLinux (https://theouterlinux.gitlab.io)         ' + Fore.CYAN + Back.CYAN + '||' + Style.RESET_ALL)
print(' ' + Fore.CYAN + Back.CYAN + '||'+ Fore.MAGENTA + Back.WHITE + ' License - GPL v3 or newer (https://www.gnu.org/licenses/)        ' + Fore.CYAN + Back.CYAN + '||' + Style.RESET_ALL)
#print (' ' + Fore.CYAN + Back.CYAN + '+-------------------------------------------------------------------++' + Style.RESET_ALL)
print(' ' + Fore.CYAN + Back.CYAN + "||"+ Fore.MAGENTA + Back.CYAN + "           Use 'q' or 'quit' any time to end program.             " + Fore.CYAN + Back.CYAN + "||" + Style.RESET_ALL)
print('')
print(Style.RESET_ALL)
def screen():
	    print (' ' + Back.CYAN + ' ' + Back.BLACK + Fore.WHITE + ' It seems as though your image is quite large (' + Fore.MAGENTA + '> '+ str(screenw) + 'x' + str(screenh) + Fore.WHITE + ').          ' + Back.CYAN + ' ' + Back.MAGENTA + ' ' + Style.RESET_ALL)
	    print (' ' + Back.CYAN + ' ' + Back.BLACK + Fore.WHITE + ' If you plan on creating with ' + Fore.CYAN + 'Brandy' + Fore.WHITE + ' and then later opening        ' + Back.CYAN + ' ' + Back.MAGENTA + ' ' + Style.RESET_ALL)
	    print (' ' + Back.CYAN + ' ' + Back.BLACK + Fore.WHITE + " with " + Fore.CYAN +"BeebEm" + Fore.WHITE + " or " + Fore.CYAN + "b-em" + Fore.WHITE + ", you will probably get a " + Fore.MAGENTA + "line is too long     " + Back.CYAN + ' ' + Back.MAGENTA + ' ' + Style.RESET_ALL)
	    print (' ' + Back.CYAN + ' ' + Back.BLACK + Fore.WHITE + ' error and the file will not load or run. So, it is ' + Fore.MAGENTA + 'suggested that ' + Back.CYAN + ' ' + Back.MAGENTA + ' ' + Style.RESET_ALL)
	    print (' ' + Back.CYAN + ' ' + Back.BLACK + Fore.WHITE + ' ' + Fore.MAGENTA + 'you shrink the file' + Fore.WHITE + ' and run this script again. However,           ' + Back.CYAN + ' ' + Back.MAGENTA + ' ' + Style.RESET_ALL)
	    print (' ' + Back.CYAN + ' ' + Back.BLACK + Fore.WHITE + ' who am I to stop you. Stretching in the following case refers     ' + Back.CYAN + ' ' + Back.MAGENTA + ' ' + Style.RESET_ALL)
	    print (' ' + Back.CYAN + ' ' + Back.BLACK + Fore.WHITE + ' to ' + Fore.MAGENTA + 'forcing' + Fore.WHITE + ' a width and height of ' + Fore.MAGENTA + screendim + Fore.WHITE + ' as opposed to keeping     ' + Back.CYAN + ' ' + Back.MAGENTA + ' ' + Style.RESET_ALL)
	    print (' ' + Back.CYAN + ' ' + Back.BLACK + Fore.WHITE + ' the original aspect ratio of the image.                           ' + Back.CYAN + ' ' + Back.MAGENTA + ' ' + Style.RESET_ALL)
	    print (' ' + Back.CYAN + '                                                                     ' + Back.MAGENTA + ' ' + Style.RESET_ALL)
	    print (' ' + Back.BLACK + ' ' + Back.MAGENTA + '                                                                     ' + Style.RESET_ALL)
	    print('')
		
#MODE 0
if modenumber in ['0','mode 0','MODE 0']:
	modenumber = '0'
	print (' ' + Back.CYAN + '                     ' + Style.RESET_ALL)
	print (' ' + Fore.MAGENTA + Back.CYAN +' ' + Back.WHITE + ' MODE 0 selected...' + Back.CYAN + ' ' + Style.RESET_ALL)
	print (' ' + Back.CYAN + '                                                                     ' + Style.RESET_ALL)
	colornumber = 2
	screenw = 75 #640
	screenh = 256
	screendim = str(screenw) + 'x' + str(screenh)
	if width > screenw or height > screenh:
		screen()
		stretchyesno = raw_input(' ' + Fore.CYAN + ' Would you like to force a ' + screendim + ' pixel image? (' + Fore.WHITE + 'y/N,q' + Fore.CYAN + '): ' + Style.RESET_ALL)
		stretchyesno = stretchyesno.lower()
		if stretchyesno in ('y', 'yes'): dimensions = '-resize ' + screendim + '\!'
		elif stretchyesno in ('q','quit','exit'):
			rmtree(tmp, ignore_errors=True)
			print('')
			exit()
		else: dimensions = '-resize ' + screendim
	else:
		stretchyesno = 'no'
		dimensions = ""
		
#MODE 1
elif modenumber in ['1', 'MODE 1', 'mode 1']:
	modenumber = '1'
	print (' ' + Back.CYAN + '                     ' + Style.RESET_ALL)
	print (' ' + Fore.MAGENTA + Back.CYAN +' ' + Back.WHITE + ' MODE 1 selected...' + Back.CYAN + ' ' + Style.RESET_ALL)
	print (' ' + Back.CYAN + '                                                                     ' + Style.RESET_ALL)
	colornumber = 4
	screenw = 75 #320
	screenh = 256
	screendim = str(screenw) + "x" + str(screenh)
	if width > screenw or height > screenh:
		screen()
		stretchyesno = raw_input(" Would you like to force a " + screendim + " pixel image? (y/N,q): ")
		stretchyesno = stretchyesno.lower()
		if stretchyesno in ('y', 'yes'): dimensions = "-resize " + screendim + "\!"
		elif stretchyesno in ('q','quit','exit'):
			rmtree(tmp, ignore_errors=True)
			print('')
			exit()
		else: dimensions = "-resize " + screendim
	else:
		stretchyesno = 'no'
		dimensions = ""
		
#MODE 2
elif modenumber in ['2', 'MODE 2', 'mode 2']:
	modenumber = '2'
	print (' ' + Back.CYAN + '                     ' + Style.RESET_ALL)
	print (' ' + Fore.MAGENTA + Back.CYAN +' ' + Back.WHITE + ' MODE 2 selected...' + Back.CYAN + ' ' + Style.RESET_ALL)
	print (' ' + Back.CYAN + '                                                                     ' + Style.RESET_ALL)
	colornumber = 8
	screenw = 75 #640
	screenh = 512
	screendim = str(screenw) + "x" + str(screenh)
	if width > screenw or height > screenh:
		screen()
		stretchyesno = raw_input(" Would you like to force a " + screendim + " pixel image? (y/N,q): ")
		stretchyesno = stretchyesno.lower()
		if stretchyesno in ('y', 'yes'): dimensions = "-resize " + screendim + "\!"
		elif stretchyesno in ('q','quit','exit'):
			rmtree(tmp, ignore_errors=True)
			print('')
			exit()
		else: dimensions = "-resize " + screendim
	else:
		stretchyesno = 'no'
		dimensions = ""
		
#MODE 4
elif modenumber in ['4','mode 4','MODE 4']:
	modenumber = '4'
	print (' ' + Back.CYAN + '                     ' + Style.RESET_ALL)
	print (' ' + Fore.MAGENTA + Back.CYAN +' ' + Back.WHITE + ' MODE 4 selected...' + Back.CYAN + ' ' + Style.RESET_ALL)
	print (' ' + Back.CYAN + '                                                                     ' + Style.RESET_ALL)
	colornumber = 2
	screenw = 75 #320
	screenh = 256
	screendim = str(screenw) + 'x' + str(screenh)
	if width > screenw or height > screenh:
		screen()
		stretchyesno = raw_input(' ' + Fore.CYAN + ' Would you like to force a ' + screendim + ' pixel image? (' + Fore.WHITE + 'y/N,q' + Fore.CYAN + '): ' + Style.RESET_ALL)
		stretchyesno = stretchyesno.lower()
		if stretchyesno in ('y', 'yes'): dimensions = '-resize ' + screendim + '\!'
		elif stretchyesno in ('q','quit','exit'):
			rmtree(tmp, ignore_errors=True)
			print('')
			exit()
		else: dimensions = '-resize ' + screendim
	else:
		stretchyesno = 'no'
		dimensions = ""
		
#MODE 5
elif modenumber in ['5', 'MODE 5', 'mode 5']:
	modenumber = '5'
	print (' ' + Back.CYAN + '                     ' + Style.RESET_ALL)
	print (' ' + Fore.MAGENTA + Back.CYAN +' ' + Back.WHITE + ' MODE 5 selected...' + Back.CYAN + ' ' + Style.RESET_ALL)
	print (' ' + Back.CYAN + '                                                                     ' + Style.RESET_ALL)
	colornumber = 4
	screenw = 75
	screenh = 256
	screendim = str(screenw) + "x" + str(screenh)
	if width > screenw or height > screenh:
		screen()
		stretchyesno = raw_input(" Would you like to force a " + screendim + " pixel image? (y/N,q): ")
		stretchyesno = stretchyesno.lower()
		if stretchyesno in ('y', 'yes'): dimensions = "-resize " + screendim + "\!"
		elif stretchyesno in ('q','quit','exit'):
			rmtree(tmp, ignore_errors=True)
			print('')
			exit()
		else: dimensions = "-resize " + screendim
	else:
		stretchyesno = 'no'
		dimensions = ""

#MODE 15
elif modenumber in ['15', 'mode 15', 'MODE 15']:
	modenumber = '15'
	print (' ' + Back.CYAN + '                      ' + Style.RESET_ALL)
	print (' ' + Fore.MAGENTA + Back.CYAN +' ' + Back.WHITE + ' MODE 15 selected...' + Back.CYAN + ' ' + Style.RESET_ALL)
	print (' ' + Back.CYAN + '                                                                     ' + Style.RESET_ALL)
	colornumber = 32
	screenw = 320
	screenh = 200
	screendim = str(screenw) + "x" + str(screenh)
	if width > screenw or height > screenh:
		screen()
		stretchyesno = raw_input(" Would you like to force a " + screendim + " pixel image? (y/N,q): ")
		stretchyesno = stretchyesno.lower()
		if stretchyesno in ('y', 'yes'): dimensions = "-resize " + screendim + "\!"
		elif stretchyesno in ('q','quit','exit'):
			rmtree(tmp, ignore_errors=True)
			print('')
			exit()
		else: dimensions = "-resize " + screendim
	else:
		stretchyesno = 'no'
		dimensions = ""
		
#QUIT
elif modenumber in ['q','quit','Q','Quit','QUIT','Exit','EXIT','exit']: exit()
	
#If invalid SCREEN option picked
else:
	modenumber = '15'
	print (' ' + Back.CYAN + '                     ' + Style.RESET_ALL)
	print (' ' + Fore.MAGENTA + Back.CYAN +' ' + Back.WHITE + ' MODE 1 selected...' + Back.CYAN + ' ' + Style.RESET_ALL)
	print (' ' + Back.CYAN + '                                                                    ' + Style.RESET_ALL)
	colornumber = 32
	screenw = 320
	screenh = 200
	screendim = str(screenw) + "x" + str(screenh)
	if width > screenw or height > screenh:
		screen()
		stretchyesno = raw_input(" Would you like to force a " + screendim + " pixel image? (y/N,q): ")
		stretchyesno = stretchyesno.lower()
		if stretchyesno in ('y', 'yes'): dimensions = "-resize " + screendim + "\!"
		elif stretchyesno in ('q','quit','exit'):
			rmtree(tmp, ignore_errors=True)
			print('')
			exit()
		else: dimensions = "-resize " + screendim
	else:
		stretchyesno = 'no'
		dimensions = ""

# Convert the original image to a limited color image for previewing
# before converting the preview image to QBASIC DATA format. GIF87 format
# was chosen for the output because it is supported by older and newer
# software where as GIF89 is mostly limited to newer software and a few
# currently maintained DOS image viewers. The image is dithered depending
# on whether or not the image has less than 256 colors before the conversion.
if colornumber == 2:
	if colorcount > colornumber:
		cmd = 'convert ' + dimensions + ' -colors 2 -depth 1 -quantize Lab +antialias -quantize Lab -remap ' + palette2 + ' "'+filepath+'"' + " GIF87:" + '"'+previewfile+'"'
		subprocess.call(cmd, shell=True)
	else:
		cmd = 'convert ' + dimensions + ' -colors 2 -depth 1 +dither +antialias -quantize Lab -remap ' + palette2 + ' "'+filepath+'"' + " GIF87:" + '"'+previewfile+'"'
		subprocess.call(cmd, shell=True)
elif colornumber == 4:
	if colorcount > colornumber:
		cmd = 'convert ' + dimensions + ' -colors 4 -depth 2 -quantize Lab +antialias -remap ' + palette4 + ' "'+filepath+'"' + " GIF87:" + '"'+previewfile+'"'
		subprocess.call(cmd, shell=True)
	else:
		cmd = 'convert ' + dimensions + ' -colors 4 -depth 2 +dither +antialias -quantize Lab -remap ' + palette4 + ' "'+filepath+'"' + " GIF87:" + '"'+previewfile+'"'
		subprocess.call(cmd, shell=True)
elif colornumber == 8:
	if colorcount > colornumber:
		cmd = 'convert ' + dimensions + ' -colors 8 -depth 3 -quantize Lab +antialias -remap ' + palette8 + ' "'+filepath+'"' + " GIF87:" + '"'+previewfile+'"'
		subprocess.call(cmd, shell=True)
	else:
		cmd = 'convert ' + dimensions + ' -colors 8 -depth 3 +dither +antialias -quantize Lab -remap ' + palette8 + ' "'+filepath+'"' + " GIF87:" + '"'+previewfile+'"'
		subprocess.call(cmd, shell=True)
elif colornumber == 32:
	if colorcount > colornumber:
		cmd = 'convert ' + dimensions + ' -colors 32 -depth 4 +antialias -remap '+ palette32 + ' "'+filepath+'"' + " GIF87:" + '"'+previewfile+'"'
		subprocess.call(cmd, shell=True)
	else:
		cmd = 'convert ' + dimensions + ' -colors 32 -depth 4 +dither +antialias -remap ' + palette32 + ' "'+filepath+'"' + " GIF87:" + '"'+previewfile+'"'
		subprocess.call(cmd, shell=True)
else:
	if colorcount > colornumber:
		cmd = 'convert ' + dimensions + ' -colors 32 -depth 4 +antialias -quantize Lab -remap ' + palette32 + ' "'+filepath+'"' + " GIF87:" + '"'+previewfile+'"'
		subprocess.call(cmd, shell=True)
	else:
		cmd = 'convert ' + dimensions + ' -colors 32 -depth 4 +dither +antialias -quantize Lab -remap ' + palette32 + ' "'+filepath+'"' + " GIF87:" + '"'+previewfile+'"'
		subprocess.call(cmd, shell=True)
		


# Ask if the user wants 'DATA' added to the beginning of each line to save
# time if the plan is to copy/pasting file contents.
clear()
print(' ' + Fore.CYAN + Back.CYAN + '+-------------------------------------------------------------------++' + Style.RESET_ALL)
print(' ' + Fore.CYAN + Back.CYAN + '||'+ Fore.MAGENTA + Back.WHITE + ' img2bbc.py - Convert a modern image to BBC BASIC DATA array      ' + Fore.CYAN + Back.CYAN + '||' + Style.RESET_ALL)
print(' ' + Fore.CYAN + Back.CYAN + '||'+ Fore.MAGENTA + Back.WHITE + ' Author - TheOuterLinux (https://theouterlinux.gitlab.io)         ' + Fore.CYAN + Back.CYAN + '||' + Style.RESET_ALL)
print(' ' + Fore.CYAN + Back.CYAN + '||'+ Fore.MAGENTA + Back.WHITE + ' License - GPL v3 or newer (https://www.gnu.org/licenses/)        ' + Fore.CYAN + Back.CYAN + '||' + Style.RESET_ALL)
#print (' ' + Fore.CYAN + Back.CYAN + '+-------------------------------------------------------------------++' + Style.RESET_ALL)
print(' ' + Fore.CYAN + Back.CYAN + "||"+ Fore.MAGENTA + Back.CYAN + "           Use 'q' or 'quit' any time to end program.             " + Fore.CYAN + Back.CYAN + "||" + Style.RESET_ALL)
print('')
print (' ' + Back.MAGENTA + '                 ' + Style.RESET_ALL)
print (' ' + Fore.MAGENTA + Back.MAGENTA +' ' + Back.WHITE + " Append 'DATA?'" + Back.MAGENTA + ' ' + Style.RESET_ALL)
print (' ' + Back.MAGENTA + '                                                                      ' + Style.RESET_ALL)
print (' ' + Back.MAGENTA + ' ' + Back.BLACK + Fore.WHITE + ' If you are planning to copy/paste the contents of the DAT file to  ' + Back.MAGENTA + ' ' + Back.CYAN + ' ' + Style.RESET_ALL)
print (' ' + Back.MAGENTA + ' ' + Back.BLACK + Fore.WHITE + " a BBC BASIC project, it may be helpful to go ahead and add 'DATA'  " + Back.MAGENTA + ' ' + Back.CYAN + ' ' + Style.RESET_ALL)
print (' ' + Back.MAGENTA + ' ' + Back.BLACK + Fore.WHITE + ' to the beginning of each line for you. This also helps if you want ' + Back.MAGENTA + ' ' + Back.CYAN + ' ' + Style.RESET_ALL)
print (' ' + Back.MAGENTA + ' ' + Back.BLACK + Fore.WHITE + ' to load the DAT file contents externally.                          ' + Back.MAGENTA + ' ' + Back.CYAN + ' ' + Style.RESET_ALL)
print (' ' + Back.MAGENTA + '                                                                      ' + Back.CYAN + ' ' + Style.RESET_ALL)
print (' ' + Back.BLACK + ' ' + Back.CYAN + '                                                                      ' + Style.RESET_ALL)
print('')
appendDATA = raw_input (Fore.CYAN + " Would you like to add 'DATA' to the beginning of each line? (" + Fore.WHITE + "Y/n,q" + Fore.CYAN + "): " + Fore.WHITE)
appendDATA = appendDATA.lower()

# Change the im variable from original to preview image	
im = Image.open(previewfile)
width, height = im.size

# To ensure RGB values will be used
rgb_im = im.convert('RGB')
pixelx = 1
pixely = 1

# Convert the preview image to a DATA array format
# Pixely is used in the for-loop before the pixelx so as to go line-by-line
# as opposed to column-by-column.
if colornumber == 0 or colornumber == 4:
	with open(bbcfilepath, "w+") as outfile:
		for pixely in range(height):
			row = []
			for pixelx in range(width):
				r,g,b = rgb_im.getpixel((pixelx, pixely))
				if [r, g, b] == [0,0,0]: row.append("0") #Black
				elif [r, g, b] == [255,0,0]: row.append("1") #Red
				elif [r, g, b] == [255,255,0]: row.append("2") #Yellow
				elif [r, g, b] == [255,255,255]: row.append("3") #White
				else: row.append("0")
			if appendDATA in ('n', 'no'): outfile.write(",".join(row) + '\n')
			elif appendDATA in ('q','quit','exit'):
				rmtree(tmp, ignore_errors=True)
				print('')
				exit()
			else: outfile.write("DATA " + ",".join(row) + '\n')
			
elif colornumber == 2 or colornumber == 8:
	with open(bbcfilepath, "w+") as outfile:
		for pixely in range(height):
			row = []
			for pixelx in range(width):
				r,g,b = rgb_im.getpixel((pixelx, pixely))
				if [r, g, b] == [0,0,0]: row.append("0") #Black
				elif [r, g, b] == [255,0,0]: row.append("1") #Red
				elif [r, g, b] == [0,255,0]: row.append("2") #Green
				elif [r, g, b] == [255,255,0]: row.append("3") #Yellow
				elif [r, g, b] == [0,0,255]: row.append("4") #Blue
				elif [r, g, b] == [255,0,255]: row.append("5") #Magenta
				elif [r, g, b] == [0,255,255]: row.append("6") #Cyan
				elif [r, g, b] == [255,255,255]: row.append("7") #white
				else: row.append("0")
			if appendDATA in ('n', 'no'): outfile.write(",".join(row) + '\n')
			elif appendDATA in ('q','quit','exit'):
				rmtree(tmp, ignore_errors=True)
				print('')
				exit()
			else: outfile.write("DATA " + ",".join(row) + '\n')
			
else:
	with open(bbcfilepath, "w+") as outfile:
		for pixely in range(height):
			row = []
			for pixelx in range(width):
				r,g,b = rgb_im.getpixel((pixelx, pixely))
				if [r, g, b] == [0,0,0]: row.append("0") #Black
				elif [r, g, b] == [68,0,0]: row.append("1") #Deep brown
				elif [r, g, b] == [68,68,0]: row.append("2") #Green brown
				elif [r, g, b] == [68,0,68]: row.append("3") #Deep violet
				elif [r, g, b] == [68,68,68]: row.append("4") #Dark gray
				elif [r, g, b] == [204,0,0]: row.append("5") #Red
				elif [r, g, b] == [204,68,0]: row.append("6") #Deep orange
				elif [r, g, b] == [204,0,68]: row.append("7") #Ruby
				elif [r, g, b] == [204,68,68]: row.append("8") #Reddish
				elif [r, g, b] == [68,136,0]: row.append("9") #Dark grass green
				elif [r, g, b] == [68,204,0]: row.append("10") #Green apple
				elif [r, g, b] == [68,136,68]: row.append("11") #Light forest green
				elif [r, g, b] == [68,204,68]: row.append("12") #Lightish green
				elif [r, g, b] == [204,136,0]: row.append("13") #Golden brown
				elif [r, g, b] == [204,204,0]: row.append("14") #Piss yellow
				elif [r, g, b] == [204,136,68]: row.append("15") #Leather
				elif r==204 and g==204 and b==68: row.append("16") #Vomit
				elif r==68 and g==0 and b==136: row.append("17") #Indigo
				elif r==68 and g==68 and b==136: row.append("18") #Twilight
				elif r==68 and g==0 and b==204: row.append("19") #Violet blue
				elif r==68 and g==68 and b==204: row.append("20") #Blue blue
				elif r==204 and g==0 and b==136: row.append("21") #Magenta
				elif r==204 and g==68 and b==136: row.append("22") #Medium pink
				elif r==204 and g==0 and b==204: row.append("23") #Purple/pink
				elif r==204 and g==68 and b==204: row.append("24") #Pinkish purple
				elif r==68 and g==136 and b==136: row.append("25") #Dark aqua
				elif r==68 and g==204 and b==136: row.append("26") #Green teal
				elif r==68 and g==136 and b==204: row.append("27") #Bluish
				elif r==68 and g==204 and b==204: row.append("28") #Topaz
				elif r==204 and g==136 and b==136: row.append("29") #Brownish pink
				elif r==204 and g==204 and b==136: row.append("30") #Khaki
				elif r==204 and g==136 and b==204: row.append("31") #Soft purple
				elif r==204 and g==204 and b==204: row.append("32") #Silver
				else: row.append("0")

        # Write row to file, joining all values with comma between them
        # and then create a new row after the end of each range of pixelx.
        # If appendDATA is 'yes', add the word 'DATA' to the beginning
        # of each line.
			if appendDATA in ('n', 'no'): outfile.write(",".join(row) + '\n')
			elif appendDATA in ('q','quit','exit'):
				rmtree(tmp, ignore_errors=True)
				print('')
				exit()
			else: outfile.write("DATA " + ",".join(row) + '\n')
			
# Move the img2bbc file to same directory as original image. This is not
# done by default in the same directory as the original image because
# it will probably freeze that directory as far as using a GUI file
# manager is concerned because there are a lot of 'echo >>' going on.
cmd = 'mv ' + bbcfilepath + ' ' + bbcfilepathfinal
subprocess.call(cmd, shell=True)

# Cleanup; deletes the temporary directory this script created at the beginning
rmtree(tmp, ignore_errors=True)
	
print(Style.RESET_ALL + "")
clear()
print(' ' + Fore.CYAN + Back.CYAN + '+-------------------------------------------------------------------++' + Style.RESET_ALL)
print(' ' + Fore.CYAN + Back.CYAN + '||'+ Fore.MAGENTA + Back.WHITE + ' img2bbc.py - Convert a modern image to BBC BASIC DATA array      ' + Fore.CYAN + Back.CYAN + '||' + Style.RESET_ALL)
print(' ' + Fore.CYAN + Back.CYAN + '||'+ Fore.MAGENTA + Back.WHITE + ' Author - TheOuterLinux (https://theouterlinux.gitlab.io)         ' + Fore.CYAN + Back.CYAN + '||' + Style.RESET_ALL)
print(' ' + Fore.CYAN + Back.CYAN + '||'+ Fore.MAGENTA + Back.WHITE + ' License - GPL v3 or newer (https://www.gnu.org/licenses/)        ' + Fore.CYAN + Back.CYAN + '||' + Style.RESET_ALL)
#print (' ' + Fore.CYAN + Back.CYAN + '+-------------------------------------------------------------------++' + Style.RESET_ALL)
print(' ' + Fore.CYAN + Back.CYAN + "||"+ Fore.MAGENTA + Back.CYAN + "           Program finished. Temporary files cleaned.             " + Fore.CYAN + Back.CYAN + "||" + Style.RESET_ALL)
print('')
print(' ' + Back.CYAN + Fore.BLACK + " Congratulations! Files created:" + Style.RESET_ALL)
print(' ' + Fore.CYAN + "Preview file: " + Style.RESET_ALL + previewfile)
print(' ' + Fore.CYAN + "BBC BASIC DATA file: " + Style.RESET_ALL + dirpath + '/' + Fore.MAGENTA + dosbase + ".DAT" + Style.RESET_ALL + "'")
print("")
print(' ' + Back.CYAN + Fore.BLACK + " If not too familiar with BBC BASIC, suggested use is:" + Style.RESET_ALL)
print(" MODE " + modenumber + "" + Style.RESET_ALL)
print(" CLS")
print(" ZOOM=7") #ZOOM=7 because without it, small DATA arrays on BBC Micros are unrecognizable
print(" STARTX=0")
print(" STARTY=0")
if stretchyesno in ('y','yes'):
	print(" FOR Y = " + str(screenh) + " TO 1 STEP -1 " + Style.RESET_ALL)
	print("   FOR X = 1 TO " + str(screenw) + " " + Style.RESET_ALL)
else:
	print(" FOR Y = " + str(height) + " TO 1 STEP -1 " + Style.RESET_ALL)
	print("   FOR X = 1 TO " + str(width) + " "  + Style.RESET_ALL)
print("     READ DotColor")
print("     GCOL 0,DotColor")
print("     MOVE X+STARTX,Y+STARTY" + Style.RESET_ALL)
print("     PLOT 65,X*ZOOM+STARTX,Y*ZOOM+STARTY")
print("   NEXT X")
print(" NEXT Y")
print("")
print(' ' + Back.CYAN + Fore.BLACK + ' Followed by:' + Style.RESET_ALL)
print(" DATA ... " + Fore.CYAN + "'Copy/paste from DAT file" + Style.RESET_ALL)
print(" DATA ... " + Fore.CYAN + "'So on and so forth..." + Style.RESET_ALL)
print("")
print(" DATA arrays in BBC BASIC are traditional placed after FOR loops and such.")

#Return colors back to normal on Windows
if os.name == 'Windows': deinit()
